!2345678911234567892123456789312345678941234567895123456789612345678971234567898
!        0         0         0         0         0         0         0         0
!+
! Module: badpix_mod
!
! Authors:  Andy Read - MPE
!          
! Description: BADPIX produces a table containing position and defect
! information of the bad pixels, and this table is attached to the input raw
! events file. 
! Comments: 
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!- 
module badpix_mod

  use cal
  use dal
  use oal
  use param  
  use errorHandling
  
  implicit none 
  
  private 

  public badpix 

contains 

  !---------------------------------------------------------------------
  subroutine badpix
  !---------------------------------------------------------------------

    implicit none

    type(CalInfo)                       :: badPixFileInfo
    type(DataSetT)                      :: in_events, in_mode3
    type(TableT)                        :: tab, tab_events, tab_exposure, tab_mode3
    type(ColumnT)                       :: xCol, yCol, typeCol, &
                                           yextCol, badflagCol, &
                                           col_mode3_rawX, col_mode3_rawY, &
                                           col_mode3_type, &
                                           col_mode3_yextent

    integer(kind=int8), pointer         :: Btype(:,:)
    integer(int16), dimension(:), allocatable :: a_x, a_y, a_type, a_yext, &
                                           a_badflag, b_x, b_y, b_type, b_yext
    integer(int16), dimension(:), pointer :: rawX, rawY, Type, yextent, &
                                           OALrawX, OALrawY, OALType, OALyextent, &
                                           p_x, p_y, p_type, p_yext, &
                                           p_badflag, p3_x, p3_y, &
                                           p3_type, p3_yext 
    integer                             :: i, ii, j, k, ccdid, bext, cext, &
                                           sum_m1, sum_m2, sum_m12, n12cal, &
                                           ntrow3, npixtot, ncalbpix, n, &
				                                       ccdnode, Nbpxfincal, Nfinal, &
                                           x0 = -1, y0 = -1, nx = -1, ny = -1, &
					                                       nuplnk, noaluplnk, &
                                           Noutwin, Ncutwin, Npurge, Nadd, &
                                           iold, xmin, ymin, xmax, ymax, x, y, &
                                           yy, pretype, revolution, &
                                           maskx0 = -1, masky0 = -1, maskdx= -1, maskdy = -1

    integer(kind=int32)                 :: calIval
    real(kind=double)                   :: calRval

    logical, dimension(:), pointer      :: uplinked, OALuplinked
    logical                             :: mode1, mode2, mode3, CalInBpxf, &
                                           allCalInBpxf, BpxfInCal, emptyext, &
                                           BpxfInCalEver, window, oal, &
                                           noCalInBpxf, mos, warnflag, &
                                           calBval, yextgot, &
					                                      delete_BADPIX = .false.

    character(len=*), parameter         :: taskname = "badpix"
    character(len=160)                  :: history
    character(len=25)                   :: datamode, tmpchar
    character(len=5)                    :: instrument
!
  character(len=12)  :: BPT_CODE     = INVALID_BAD_PIXEL_TABLE_CODE, &
 		                     BPT_CODE_CCF = INVALID_BAD_PIXEL_TABLE_CODE
!                       BPT_CODE = '0_000_UNK', BPT_CODE_CCF = '0_000_UNK'
  integer(kind=Int16):: BPT_CODE_MIN_REV = 293    ! first rev of implementation

!******************************************************************************
!                 START PROGRAM
!******************************************************************************
!------------------------------------------------------------
!	GET INPUT EVENTS
!------------------------------------------------------------
    if (booleanParameter("withoutset")) then 
      in_events = clone(stringParameter("eventset"), &
                  stringParameter("outset"), MODIFY, USE_ENVIRONMENT)
    else
      in_events = dataSet(stringParameter("eventset"),MODIFY)
    endif
!------------------------------------------------------------
!	GET NAME OF SECONDARY EXTENSION and DATAMODE 
!------------------------------------------------------------
    if (hasBlock(in_events,"EVENTS")) then	!	EPN (all modes)
      tab_events = table(in_events,"EVENTS") 
    elseif (hasBlock(in_events,"M1IME1")) then	!       EMOS1 (imaging)
      tab_events = table(in_events,"M1IME1")
      datamode = "IMAGING"
    elseif (hasBlock(in_events,"M1TIE1")) then	!       EMOS1 (timing)
      tab_events = table(in_events,"M1TIE1")
      datamode = "TIMING"
    elseif (hasBlock(in_events,"M1RIE1")) then	!       EMOS1 (reduced imaging)
      tab_events = table(in_events,"M1RIE1")
      datamode = "REDUCED IMAGING"
    elseif (hasBlock(in_events,"M1CTE1")) then	!       EMOS1 (compressed timing)
      tab_events = table(in_events,"M1CTE1")
      datamode = "COMPRESSED TIMING"
    elseif (hasBlock(in_events,"M2IME1")) then	!       EMOS2 (imaging)
      tab_events = table(in_events,"M2IME1")
      datamode = "IMAGING"
    elseif (hasBlock(in_events,"M2TIE1")) then	!       EMOS2 (timing)
      tab_events = table(in_events,"M2TIE1")
      datamode = "TIMING"
    elseif (hasBlock(in_events,"M2RIE1")) then	!       EMOS2 (reduced imaging)
      tab_events = table(in_events,"M2RIE1")
      datamode = "REDUCED IMAGING"
    elseif (hasBlock(in_events,"M2CTE1")) then	!       EMOS2 (compressed timing)
      tab_events = table(in_events,"M2CTE1")
      datamode = "COMPRESSED TIMING"
    else 
      call error("badExtension",&
              "No, invalid or incorrectly named extension in input events file")
    endif

    if (hasBlock(in_events,"EXPOSURE")) then
      tab_exposure = table(in_events,"EXPOSURE") 
      if (hasAttribute(tab_exposure,"MASKX0")) maskx0=int16Attribute(tab_exposure,"MASKX0")
      if (hasAttribute(tab_exposure,"MASKY0")) masky0=int16Attribute(tab_exposure,"MASKY0")
      if (hasAttribute(tab_exposure,"MASKDX")) maskdx=int16Attribute(tab_exposure,"MASKDX")
      if (hasAttribute(tab_exposure,"MASKDY")) maskdy=int16Attribute(tab_exposure,"MASKDY")      
    end if
!--------------------------------------------------------------------
!	GET emptyext and window
!--------------------------------------------------------------------
    emptyext = booleanParameter("emptyextension")
    window = booleanParameter("windowfilter")

!--------------------------------------------------------------------
!	SET CAL 
!--------------------------------------------------------------------
    mos = .false.
    call CAL_setState(in_events)
    call CAL_setState(accuracyLevel=1, randomize=.FALSE.)

    call CAL_getState(instrument=calIval)
    write (errstr,*) "Instrument         = ",instId(calIval)
    call message(AppMsg,Verbose)
    instrument = instId(calIval)
    if (instrument == "EMOS1" .or. instrument == "EMOS2") then
      mos = .true.
    elseif (instrument == "EPN") then
      mos = .false.
      if (hasAttribute(tab_events,"F1128AVE")) then  ! CCD = -90C
        calRval = real64Attribute(tab_events,"F1128AVE")+273.15
	       if (calRval < 0) then   ! do nothing
	         calRval = 183.2
	       else
          call CAL_setState(temperature = calRval)
          write (errstr,'(a,f8.3,a)') " EPN CCD temperature set to",calRval,' [K]'
          call message(AppMsg,Noisy)
        endif
      endif
      if (hasAttribute(tab_events,"TQB1_AVE")) then  ! quadrant box = +21C
        calRval = real64Attribute(tab_events,"TQB1_AVE")+273.15
	       if (calRval < 0) then   ! do nothing
	         calRval = 294.0
	       else
          call CAL_setState(temperatureCamera = calRval)
          write (errstr,'(a,f8.3,a)') ' EPN box temperature set to',calRval,' [K]'
          call message(AppMsg,Noisy)
        endif
      endif
    else
      mos = .false.
    endif

    call CAL_getState(ccdChipId=calIval)
    write (errstr,*) "CCD chip           = ",calIval
    call message(AppMsg,Verbose)
    ccdid = calIval
    if (mos) then
      call CAL_getState(ccdNodeId=calIval)
      write (errstr,*) "CCD node           = ",calIval
      call message(AppMsg,Verbose)
      ccdnode = calIval
    endif
    call CAL_getState(ccdModeId=calIval)
    write (errstr,*) "CCD mode           = ",modeId(calIval)
    call message(AppMsg,Verbose)
    call CAL_getState(filterId=calIval)
    write (errstr,*) "Filter             = ",filterId(calIval)
    call message(AppMsg,Verbose)
    call CAL_getState(accuracyLevel=calIval)
    write (errstr,*) "Accuracy level     = ",calIval
    call message(AppMsg,Verbose)
    call CAL_getState(temperature=calRval)
    write (errstr,*) "Temperature        = ",calRval
    call message(AppMsg,Verbose)
    call CAL_getState(temperatureCamera=calRval)
    write (errstr,*) "Camera Temperature = ",calRval
    call message(AppMsg,Verbose)
    call CAL_getState(randomize=calBval)
    write (errstr,*) "Randomization      = ",calBval
    call message(AppMsg,Verbose)
!------------------------------------------------------------
!	TEST INSTRUMENT (Secondary header)
!------------------------------------------------------------
!   instrument = stringAttribute(tab_events,"INSTRUME")
    if (instrument /= "EPN" .and. &
        instrument /= "EMOS1" .and. &
        instrument /= "EMOS2") then
       write(errstr,*) "Instrument not supported: ",trim(instrument)
       call error("badInstrument")
    endif
!    if (instrument == "EPN") then
!       ccdid = 1 + (3 * int8Attribute(tab_events,"QUADRANT")) + &
!                        int8Attribute(tab_events,"CCDID")
!     endif
!     write (errstr,*) "Datamode: ",datamode
!     call message(AppMsg,Verbose)

!------------------------------------------------------------
!	get window keywords
!------------------------------------------------------------

      x0 = int16Attribute(tab_events,"WINDOWX0")
      y0 = int16Attribute(tab_events,"WINDOWY0")
      nx = int16Attribute(tab_events,"WINDOWDX")
      ny = int16Attribute(tab_events,"WINDOWDY")

!------------------------------------------------------------
!	DOES EXTENSION EXIST? 
!------------------------------------------------------------
    if (hasBlock(in_events,"BADPIX")) then	!	BADPIX extension exists
      call warning("overwriteBadpix","BADPIX extension already exists - overwriting")
      delete_BADPIX = .true.
    endif
!------------------------------------------------------------
!	If (PN?) timing/burst - empty extension
!------------------------------------------------------------
  if (emptyext) then
    write (errstr,*) "emptyextension=Y chosen, added extension is empty"
    call message(AppMsg,Verbose)
  else					!OK - normal badpix
!------------------------------------------------------------
!	GET BADPIX MODES (1,2,3) ?
!------------------------------------------------------------
    mode1 = booleanParameter("getuplnkbadpix")
    mode2 = booleanParameter("getotherbadpix")
    mode3 = booleanParameter("getnewbadpix")
    if (mode1) call message(AppMsg,Verbose,"Mode1 - uplinked bad pixels (in CCF)")
    if (mode2) call message(AppMsg,Verbose,"Mode2 - non-uplinked bad pixels (in CCF)")
    if (mode3) call message(AppMsg,Verbose,"Mode3 - new bad pixels (from BADPIXFIND)")
    if (.not.mode1 .and. .not.mode2 .and. .not.mode3) then       
      call error("noBadpixModes","No modes chosen - no bad pixels added")
    endif
!--------------------------------------------------------------------
!	CALL CAL BADPIX 
!--------------------------------------------------------------------
    call CAL_ccfInfo("BADPIX", badPixFileInfo)
!--------------------------------------------------------------------
!       check EPN_BADPIX.CCF: read BPT_CODE from events and CCF files
!--------------------------------------------------------------------
    if (instrument == "EPN") then
      if (hasAttribute(tab_events,"REVOLUT")) then
        revolution = int16Attribute(tab_events,"REVOLUT")
      elseif (hasAttribute(tab_events,"ODFLATEX")) then
        tmpchar = trim(stringAttribute(tab_events,"ODFLATEX"))
        read(tmpchar(1:4),'(i4.4)') revolution
       	else
        revolution = -1 
      endif

      if (hasAttribute(tab_events,"BPT_CODE")) then
        BPT_CODE = trim(stringAttribute(tab_events,"BPT_CODE"))
      else
        BPT_code = INVALID_BAD_PIXEL_TABLE_CODE
      endif

      BPT_CODE_CCF = CAL_getBadPixelTableCode()
!      print *,"BPT:",BPT_CODE_CCF,INVALID_BAD_PIXEL_TABLE_CODE(1:9) 
      if (verbosity(AppMsg) > Verbose) then
        write(errstr,*) 'BPT_CODE_CCF = ',BPT_CODE_CCF(1:9)
	       call message(AppMsg,Sparse)
        write(errstr,*) 'BPT_CODE     = ',BPT_CODE(1:9)
	       call message(AppMsg,Sparse)
        write(errstr,*) 'BPT_CODE_UNK = ',INVALID_BAD_PIXEL_TABLE_CODE(1:9)
	       call message(AppMsg,Sparse)
      endif
      if (REVOLUTION < BPT_CODE_MIN_REV) then
        BPT_CODE = INVALID_BAD_PIXEL_TABLE_CODE
        write(errstr,'(a,i3,a,i3,a)') &
	           'BadPixelTable code not implemented before revolution ', &
             BPT_CODE_MIN_REV,' [here Rev = ',revolution,']'
        call message(AppMsg,Verbose)
      elseif (BPT_CODE_CCF(1:9) == INVALID_BAD_PIXEL_TABLE_CODE(1:9)) then
        write(errstr,*) 'BadPixelTable code ',BPT_CODE(1:12), &
                    ' is not coded in BADPIX.CCF'
        call message(AppMsg,Sparse)
      elseif (BPT_CODE_CCF(1:9) .ne. BPT_CODE(1:9)) then
        write(errstr,*) 'BadPixelTable code differs in PAH.FITS [', &
          BPT_CODE(1:9),'] and BADPIX.CCF [',BPT_CODE_CCF(1:9),']'
        call warning('BPTcode')
        call message(AppMsg,Sparse,' ')
      else
        write(errstr,*) 'BadPixelTable code: ',BPT_CODE(1:12)
        call message(AppMsg,Sparse)
      endif   
    endif
!------------------------------------------------------------
!	FIND SIZE OF MODE3 FILE (badpixfind file)
!------------------------------------------------------------
    if (mode3) then
      in_mode3 = dataSet(stringParameter("badpixset"),READ)
      tab_mode3 = table(in_mode3,"BADPIX")
      ntrow3 = numberOfRows(tab_mode3)
!------------------------------------------------------------
!       CHECK IF BADPIXFIND FILE IS FOR CORRECT INSTRUMENT
!------------------------------------------------------------
      if (stringAttribute(tab_events,"INSTRUME") .ne. &
          stringAttribute(tab_mode3,"INSTRUME") .and. mode3) then
        call warning("badBadpixfindFile","badpixfind file is incompatible with input file (INSTRUME) - Mode 3 not accessed")
        mode3 = .FALSE.
        ntrow3 = 0
      endif
!------------------------------------------------------------
!       CHECK IF BADPIXFIND FILE IS FOR CORRECT CCD/NODE
!------------------------------------------------------------
      if (instrument == "EPN" .and. mode3) then 
        if (int8Attribute(tab_events,"QUADRANT") /= &
          int8Attribute(tab_mode3,"QUADRANT") .or. &
          int8Attribute(tab_events,"CCDID") /= &
          int8Attribute(tab_mode3,"CCDID")) then
          call warning("badBadpixfindFile","badpixfind file is incompatible with input file (QUADRANT/CCDID) - Mode 3 not accessed")
          mode3 = .FALSE.
          ntrow3 = 0
        endif
      elseif (mos) then
        ccdnode = int8Attribute(tab_mode3,"CCDNODE")
        if (int8Attribute(tab_events,"CCDID") /= &
          int8Attribute(tab_mode3,"CCDID") .and. mode3) then
          call warning("badBadpixfindFile","badpixfind file is incompatible with input file (CCDID) - Mode 3 not accessed")
          mode3 = .FALSE.
          ntrow3 = 0
        endif
        if (int8Attribute(tab_events,"CCDNODE") /= &
          int8Attribute(tab_mode3,"CCDNODE") .and. mode3) then
          call warning("badBadpixfindFile","badpixfind file is incompatible with input file (CCDNODE) - Mode 3 not accessed")
          mode3 = .FALSE.
          ntrow3 = 0
        endif
        if (.not.mos) then                        !right thing to do???
          if (int8Attribute(tab_events,"CCDOCB") /= &
            int8Attribute(tab_mode3,"CCDOCB") .and. mode3) then
            call warning("badBadpixfindFile","badpixfind file is incompatible with input file (CCDOCB) - Mode 3 not accessed")
            mode3 = .FALSE.
            ntrow3 = 0
          endif
        endif 
      endif
    else
      ntrow3 = 0
    endif
!------------------------------------------------------------
!	GET BAD PIXEL INFO (OAL for MOS)
!------------------------------------------------------------
    call message(AppMsg,Verbose,"Getting bad pixel tables")
    call CAL_setState(ccdChipId = ccdid)
    oal=.FALSE. 
    if (mos) then 
      call OAL_setState(instId(instrument))
      call OAL_setState(ccdNr=ccdid)
      if (OAL_getBadPixelList(OALrawX, OALrawY, OALType, OALyextent, &
                              OALuplinked)) then
!        write (*, *) instrument,' chip #', ccdid, ' has ', &
!                     size(OALrawX), ' bad pixels:'
!        write (*, *) '   x    y   defect  y-extent uplinked'
!        do i=0, size(OALrawX)-1
!             write(*, '(I5,I5,I9,I10,5X,L1)') OALrawX(i), OALrawY(i), &
!                   OALType(i), OALyextent(i), OALuplinked(i)
!        enddo
        oal=.TRUE.
      else
        oal=.FALSE.  
        call message(AppMsg,Verbose,"The ODF contains no bad pixel tables")
      endif
    endif
!------------------------------------------------------------
!	GET BAD PIXEL INFO (CCF for PN/MOS)
!------------------------------------------------------------
    call CAL_getBadPixelList(rawX, rawY, Type, yextent,uplinked)
    ncalbpix = size(rawX)
    write (errstr,*) "CCF contains ",ncalbpix," bad pixel sets"
    call message(AppMsg,Verbose)
!------------------------------------------------------------
!	MOS and OAL - fill uplinked with OAL uplinked 
!------------------------------------------------------------
    if (mos .and. oal) then 
      noaluplnk = 0
      nuplnk = 0
      do i=0, size(OALrawX)-1
        if (OALuplinked(i)) noaluplnk = noaluplnk + 1
      enddo
      do i=0, size(rawX)-1
        if (uplinked(i)) nuplnk = nuplnk + 1
      enddo
    endif

!    write (*,*) nuplnk, noaluplnk
!------------------------------------------------------------
!	FIND SIZE OF MODE1/2 TABLE TO BE CREATED (n12cal)
!------------------------------------------------------------	
    n12cal=0
    sum_m1=0
    sum_m2=0
    sum_m12=0
    if (ncalbpix >= 1) then
      if (mos .and. oal) then 
        do i=0, size(OALrawX)-1
          if (mode1 .and. OALuplinked(i)) then 
            sum_m1 = sum_m1 + 1
          endif
        enddo
      else       
        do i=0, size(rawX)-1
          if (mode1 .and. uplinked(i)) then 
            sum_m1 = sum_m1 + 1
          endif
        enddo
      endif
      do i=0, size(rawX)-1
        if (mode2 .and. .not.uplinked(i)) then
          sum_m2 = sum_m2 + 1
        endif
      enddo
      n12cal = sum_m1 + sum_m2
    endif
    if (mode1) then
      write (errstr,*) sum_m1," mode1 pixel sets (uplinked to the satelite)"
      call message(AppMsg,Verbose)
      if (oal) call message(AppMsg,Verbose," - Contained within ODF files")
    endif
    if (mode2) then
      write (errstr,*) sum_m2," mode2 pixel sets (not uplinked to the satelite)"
      call message(AppMsg,Verbose)
    endif
    if (mode3 .and. .not.mode1 .and. .not.mode2) then 	! i.e. just mode3
      n12cal=0
    endif
!------------------------------------------------------------
!	ADD MODE3 ROWS TO n12cal TO GET npixtot
!------------------------------------------------------------
    npixtot=n12cal+ntrow3	
    if (mode3) then 
      write (errstr,*) ntrow3," pixel sets contained within badpixfind file"
      call message(AppMsg,Verbose)
    endif
    write (errstr,*) npixtot," pixel sets in total"
    call message(AppMsg,Verbose)
!------------------------------------------------------------
!       ADD MASK WINDOW to npixtot    
!------------------------------------------------------------
    if (instrument=="EPN".and.ccdid==4.and.maskdx>0) npixtot=npixtot+maskdx
!------------------------------------------------------------
!	ALLOCATE EXTENSION
!------------------------------------------------------------
    i = npixtot + 100 !might need to split some bpxfs in purge
    allocate (a_x(i),a_y(i),a_yext(i),a_type(i),a_badflag(i))
!------------------------------------------------------------
!       COPY MASK PIXELS TO BP EXTENSION
!------------------------------------------------------------
    n=1
    if (instrument=="EPN".and.ccdid==4.and. &
        maskx0>0.and.masky0>0.and.maskdx>0.and.maskdy>0) then
        call message(AppMsg,Sparse,"Central CCD: Writing MASK mode pixels to BADPIX extension")
      do i=maskx0,maskx0+maskdx-1
        a_x(n) = i
        a_y(n) = masky0
        a_type(n)= 0
        a_yext(n)= maskdy
        a_badflag(n)=1
        n=n+1
      end do
    end if
!------------------------------------------------------------
!	LOOP THROUGH EACH BAD CAL PIXEL AND FILL EXTENSION (MODES 1+2)
!------------------------------------------------------------		
    call message(AppMsg,Noisy,"111   X    Y  TYPE  Y-EXTENT  MODE")
    if (mode1) then
      if (ncalbpix >= 1) then 
        if (oal) then
          do i=0, size(OALrawX)-1	!loop through main node OAL uplinked
            if (mode1 .and. OALuplinked(i)) then 
              a_x(n) = OALrawX(i)
              a_y(n) = OALrawY(i)
              a_type(n) = OALType(i) 
              a_yext(n) = OALyextent(i)
              a_badflag(n) = 1
              write (errstr, "(I5,I5,I5,I8,I8)") a_x(n),a_y(n),a_type(n),a_yext(n),a_badflag(n)            
              call message(AppMsg,Noisy)
              n = n+1
            endif
          enddo
        else
          do i=0, size(rawX)-1		!loop through main node CCF uplinked
            if (mode1 .and. uplinked(i)) then 
              a_x(n) = rawX(i)
              a_y(n) = rawY(i)
              a_type(n) = Type(i) 
              a_yext(n) = yextent(i)
              a_badflag(n) = 1
              write (errstr, "(I5,I5,I5,I8,I8)") a_x(n),a_y(n),a_type(n),a_yext(n),a_badflag(n)            
              call message(AppMsg,Noisy)
              n = n+1
            endif
          enddo
        endif
      endif   
    endif
    if (mode2) then
      if (ncalbpix >= 1) then 
        do i=0, size(rawX)-1			!loop through main node CCF non-uplinked
          if (mode2 .and. .not.uplinked(i)) then 
            a_x(n) = rawX(i)
            a_y(n) = rawY(i)
            a_type(n) = Type(i) 
            a_yext(n) = yextent(i)
            a_badflag(n) = 2
            write (errstr, "(I5,I5,I5,I8,I8)") a_x(n),a_y(n),a_type(n),a_yext(n),a_badflag(n)            
            call message(AppMsg,Noisy)
            n = n+1
          endif
        enddo
      endif   
    endif
!------------------------------------------------------------		
!	SAME FOR MODE3 + checking/purging (only if mode3 + >1)
!------------------------------------------------------------		
    Npurge = 0
    if (mode3 .and. ntrow3 >= 1) then
      col_mode3_rawX = column(tab_mode3,"RAWX",READ)
      col_mode3_rawY = column(tab_mode3,"RAWY",READ)
      col_mode3_type = column(tab_mode3,"TYPE",READ)
      col_mode3_yextent = column(tab_mode3,"YEXTENT",READ)
	!      col_mode3_badflag = column(tab_mode3,"BADFLAG",READ)
      p3_x => Int16Data(col_mode3_rawX)
      p3_y => Int16Data(col_mode3_rawY)
      p3_type => Int16Data(col_mode3_type)
      p3_yext => Int16Data(col_mode3_yextent)
	!      p3_badflag =>  Int16Data(col_mode3_badflag)
      !------------------------------------------------------------		
      !	Check badpixfind entries for doubles ???	hot/dead/flick
      !------------------------------------------------------------		
      xmin = 1300
      xmax = -1
      ymin = 1300
      ymax = -1
      do i = 0, ntrow3-1
        if (p3_x(i) < xmin) xmin = p3_x(i)  
        if (p3_x(i) > xmax) xmax = p3_x(i)  
        if (p3_y(i) < ymin) ymin = p3_y(i)  
        if ((p3_y(i)+p3_yext(i)) > ymax) ymax = p3_y(i)+p3_yext(i)-1  
      enddo
      allocate(Btype(xmin:xmax,ymin:ymax))
      do x = xmin,xmax
        do y = ymin,ymax
          Btype(x,y) = 0
        enddo
      enddo

      do i = 0, ntrow3-1
        if (p3_type(i) == 1) then 
          do y = 1,p3_yext(i)
            Btype(p3_x(i),p3_y(i)+y-1) = 1
          enddo
        elseif (p3_type(i) == 3) then 
          do y = 1,p3_yext(i)
            if (Btype(p3_x(i),p3_y(i)+y-1) /= 1) then 
              Btype(p3_x(i),p3_y(i)+y-1) = 3
            endif
          enddo
        elseif (p3_type(i) == 2) then 
          do y = 1,p3_yext(i)
            if (Btype(p3_x(i),p3_y(i)+y-1) /= 1 .and. &
                Btype(p3_x(i),p3_y(i)+y-1) /= 3) then 
              Btype(p3_x(i),p3_y(i)+y-1) = 2
            endif
          enddo
        elseif (p3_type(i) == 4) then 
          do y = 1,p3_yext(i)
            if (Btype(p3_x(i),p3_y(i)+y-1) /= 1 .and. &
                Btype(p3_x(i),p3_y(i)+y-1) /= 3 .and. &
                Btype(p3_x(i),p3_y(i)+y-1) /= 2) then 
              Btype(p3_x(i),p3_y(i)+y-1) = 4
            endif
          enddo
        endif
      enddo

      allocate(b_x(0:2*ntrow3),b_y(0:2*ntrow3),b_yext(0:2*ntrow3),b_type(0:2*ntrow3))
      i=0
      do x = xmin,xmax
        pretype=-1
        do y = ymin,ymax
          if (Btype(x,y) /= pretype) then !change
            if (Btype(x,y) /= 0) then     !bpxf
              b_x(i) = x
              b_y(i) = y            
              b_type(i) = Btype(x,y)
              pretype = Btype(x,y)
              yextgot = .FALSE.
              if (y == ymax) then 
                b_yext(i) = 1
              else
                do yy = 1,ymax-y
                  if (Btype(x,y+yy) /= pretype .and. .not.yextgot) then 
                    b_yext(i) = yy
                    yextgot = .TRUE. 
                  endif
                  if (yy == ymax-y .and. .not.yextgot) b_yext(i) = yy+1
                enddo
              endif
              i = i+1
            else
              pretype=0
            endif
          endif
        enddo
      enddo

      if (i /= ntrow3) then 
        call message(AppMsg,Sparse,"Note - Badpixfind double entries found: purging badpixfind data")
        write(errstr,*) ntrow3," badpixfind entries converted to ",i," entries"
        call message(AppMsg,Verbose)
        ntrow3 = i
        npixtot=n12cal+ntrow3	
      endif
!      do i = 0,ntrow3-1
!      write (*,*) b_x(i),b_y(i),b_yext(i),b_type(i)
!      enddo
      do i=0,ntrow3-1
        a_x(n) = b_x(i)
        a_y(n) = b_y(i)
        a_type(n) = b_type(i)
        a_yext(n) = b_yext(i)
        a_badflag(n) = 3
        n = n+1
        if (.not.mode1 .and. .not.mode2) then 
          write (errstr, "(I5,I5,I5,I8,I8)") b_x(i),b_y(i),b_type(i),b_yext(i),3
          call message(AppMsg,Noisy)
        endif
      enddo
      a_badflag(n) = 3
      !------------------------------------------------------------		
      !	Check badpixfind file against CAL (still if mode3 and > 1) 
      !------------------------------------------------------------		
      allCalInBpxf = .TRUE.
      noCalInBpxf = .TRUE.
      if (size(rawX) >= 1) then 
        do j = 0, size(rawX)-1
          CalInBpxf = .FALSE.
          do i = 0, ntrow3-1		!??? probs with bothnodes nodes unequal
            if (rawX(j) .eq. b_x(i) .and. rawY(j) .eq. b_y(i) .and. &
                Type(j) .eq. b_type(i) .and. yextent(j) .eq. b_yext(i)) then 
              CalInBpxf = .TRUE.
            endif
          enddo
          if (CalInBpxf .and. allCalInBpxf) then 
            allCalInBpxf = .TRUE.
          else
            allCalInBpxf = .FALSE.
          endif
          if (.not.CalInBpxf .and. noCalInBpxf) then 
            noCalInBpxf = .TRUE.
          else
            noCalInBpxf = .FALSE.
          endif
        enddo
        if (noCalInBpxf) then
          call message(AppMsg,sparse,"Note - No exact CCF entries contained within Badpixfind file ")
        else if (.not.allCalInBpxf .and. .not.noCalInBpxf) then
          call message(AppMsg,sparse,"Note - Not all exact CCF entries contained within Badpixfind file ")
        else
          call message(AppMsg,sparse,"Note - All exact CCF entries contained within Badpixfind file ")
        endif
      endif
    endif
    !------------------------------------------------------------
    !	Purge mode3 pixels (if in cal)
    !------------------------------------------------------------
    Nbpxfincal = 0
    BpxfInCalEver = .FALSE.
    if (mode3 .and. npixtot > ntrow3) then  !mode3 and cal 
      do i = n12cal+1, npixtot              !loop through bpxf pixels
        ii = i - Nbpxfincal
        BpxfInCal = .FALSE.
        do j = 1, n12cal		    !loop through cal pixels
          if (a_x(ii) .eq. a_x(j) .and. a_y(ii) .eq. a_y(j) .and. &
              a_type(ii) .eq. a_type(j) .and. a_yext(ii) .eq. a_yext(j)) then 
            BpxfInCal = .TRUE.        
            BpxfInCalEver = .TRUE.
          endif
        enddo
        if (BpxfInCal) then 
          do k = i-Nbpxfincal, npixtot-1
            a_x(k) = a_x(k+1)
            a_y(k) = a_y(k+1)
            a_type(k) = a_type(k+1)
            a_yext(k) = a_yext(k+1)
            a_badflag(k) = a_badflag(k+1)
          enddo
          Nbpxfincal = Nbpxfincal + 1
        else
          write (errstr, "(I5,I5,I5,I8,I8)") a_x(ii),a_y(ii),a_type(ii),a_yext(ii),3
          call message(AppMsg,Noisy)
        endif
      enddo
      if (BpxfInCalEver) then 
        write(errstr,*) "Note - ",Nbpxfincal," of ",ntrow3," badpixfind entries in CCF: purging extension"
        call message(AppMsg,Sparse)
        write(errstr,*) ntrow3-Nbpxfincal," badpixfind entries remaining "
        call message(AppMsg,Verbose)
      endif
    endif
    !-----------------------------------------------------------------
    !	Purge/combine remaining mode3 pixels (if overlap with cal)
    !-----------------------------------------------------------------
    Nadd = 0
    if (mode3 .and. npixtot > ntrow3) then  !mode3 and cal  
      warnflag = .FALSE.
      do i = n12cal+1, npixtot-Nbpxfincal   !loop through remaining bpxf pixels
        iold = 0
        do bext = 0, a_yext(i)-1            !loop through every point of bpxf
          do j = 1, n12cal		    !loop through cal pixels
            do cext = 0,a_yext(j)-1         !loop through every point of cal
              if (a_x(i) == a_x(j) .and. a_y(i)+bext == a_y(j)+cext .and. & 
                  a_yext(i) >= 1) then 
                warnflag = .TRUE.
                if (a_type(i) == a_type(j)) then !same type
                  write(errstr,*) "common CCF/badpixfind pixel (same type): X ",a_x(i)," Y ",a_y(i)+bext," - Purging"
                else                             !different type
                  write(errstr,*) "common CCF/badpixfind pixel (different type): X ",a_x(i)," Y ",a_y(i)+bext," - Purging"
                endif
                call message(AppMsg,Noisy)

                if (a_y(i) >= a_y(j)) then 
                  a_y(i) = a_y(i)+1
                  a_yext(i) = a_yext(i)-1

                else				!split into two entries
                  if (iold == 0) then 
                    Nadd = Nadd + 1
                    a_x(npixtot-Nbpxfincal+Nadd) = a_x(i)
                    a_y(npixtot-Nbpxfincal+Nadd) = a_y(j)+1
                    a_type(npixtot-Nbpxfincal+Nadd) = a_type(i)
                    a_yext(npixtot-Nbpxfincal+Nadd) = a_yext(i) - (a_y(j) - a_y(i)) - 1
                    a_yext(i) = a_y(j) - a_y(i)   !
                  endif 
                  if (iold == 0) iold = i
                endif
              endif
            enddo
          enddo
        enddo
      enddo
      if (Nadd > 0) then 			             !go through added end BPXFs
        do i = npixtot-Nbpxfincal+1, npixtot-Nbpxfincal+Nadd !loop through added bpxf pixels
          do bext = 0, a_yext(i)-1                           !loop through every point of bpxf
            do j = 1, n12cal		                     !loop through cal pixels
              do cext = 0,a_yext(j)-1                        !loop through every point of cal
                if (a_x(i) == a_x(j) .and. a_y(i)+bext == a_y(j)+cext .and. & 
                    a_yext(i) >= 1) then 
                  warnflag = .TRUE.
                  if (a_type(i) == a_type(j)) then !same type
                    write(errstr,*) "common CCF/badpixfind pixel (same type): X ",a_x(i)," Y ",a_y(i)+bext," - Purging"
                  else                             !different type
                    write(errstr,*) "common CCF/badpixfind pixel (different type): X ",a_x(i)," Y ",a_y(i)+bext," - Purging"
                  endif
                  call message(AppMsg,Noisy)
                  a_y(i) = a_y(i)+1
                  a_yext(i) = a_yext(i)-1
                endif
              enddo
            enddo
          enddo
        enddo
      endif
      do i = n12cal+1, npixtot-Nbpxfincal+Nadd   !loop through remaining bpxf pixels
        if (a_yext(i-Npurge) < 1) then                  !entry vanishes
          do j = i-Npurge, npixtot-Nbpxfincal+Nadd-Npurge
            a_x(j) = a_x(j+1)
            a_y(j) = a_y(j+1)
            a_type(j) = a_type(j+1)
            a_yext(j) = a_yext(j+1)
            a_badflag(j) = a_badflag(j+1)
          enddo
          Npurge = Npurge + 1
        endif
      enddo
      if (warnflag) then 
        call message (AppMsg,Sparse,"Note - Common CCF/badpixfind pixels found with non-identical entries - Purging")
      endif 
    endif  		!END of badpixfind mode3/cal access/check/purge
    !-----------------------------------------------------------------
    !	Window mode? Only give relevant bad pixels
    !-----------------------------------------------------------------
    Noutwin = 0
    Ncutwin = 0
!   x0 = int16Attribute(tab_events,"WINDOWX0")
!   y0 = int16Attribute(tab_events,"WINDOWY0")
!   nx = int16Attribute(tab_events,"WINDOWDX")
!   ny = int16Attribute(tab_events,"WINDOWDY")
    if (window) then 
      !-------------------------------------------------------------
      !   GET WINDOW ORIGIN/SIZE
      !-------------------------------------------------------------
      write(errstr,*) "Window origin at RAWX= ",x0,", RAWY= ",y0
      call message(AppMsg,Verbose)
      write(errstr,*) "Window size in RAWX= ",nx,", in RAWY= ",ny
      call message(AppMsg,Verbose)
      do i = 1, npixtot-Nbpxfincal+Nadd-Npurge
        warnflag = .false. 
        if (a_x(i-Noutwin) < x0 .or. a_x(i-Noutwin) >= x0+nx .or. &
!           a_y(i-Noutwin)+a_yext(i-Noutwin) < y0 .or. &
            a_y(i-Noutwin)+a_yext(i-Noutwin)-1 < y0 .or. &
            a_y(i-Noutwin) >= y0+ny ) then !out of window range
          do j = i-Noutwin, npixtot-Nbpxfincal-Noutwin-Npurge+Nadd
            a_x(j) = a_x(j+1)
            a_y(j) = a_y(j+1)
            a_type(j) = a_type(j+1)
            a_yext(j) = a_yext(j+1)
            a_badflag(j) = a_badflag(j+1)
          enddo
          Noutwin = Noutwin + 1
        else			! OK/cut
          if (a_y(i-Noutwin) < y0) then 	!bottom
            a_yext(i-Noutwin) = a_yext(i-Noutwin) - (y0 - a_y(i-Noutwin))
            a_y(i-Noutwin) = y0
            Ncutwin = Ncutwin +1
            warnflag = .true.
          endif
          if (a_y(i-Noutwin)+a_yext(i-Noutwin) >= y0+ny) then 	!top
            a_yext(i-Noutwin) = a_yext(i-Noutwin) - &
                  ( (a_y(i-Noutwin)+a_yext(i-Noutwin)) - (y0+ny) )
            if (.not.warnflag) Ncutwin = Ncutwin +1
          endif
        endif
      enddo
      if (Noutwin > 0) then
        write (errstr,*) Noutwin," pixel sets out of WINDOW range - deleting"
        call message(AppMsg,Verbose)
      endif
      if (Ncutwin > 0) then
        write (errstr,*) Ncutwin," pixel sets overlapping WINDOW range - altering"
        call message(AppMsg,Verbose)
      endif
    endif
!------------------------------------------------------------
!	end of emptyext if below:
!------------------------------------------------------------
  endif
!------------------------------------------------------------
!	ADD AN EXTENSION OF SIZE npixtot (might get smaller)
!------------------------------------------------------------
    if (emptyext) then 
    !----
    !  empty memory allocation. Only in case emptyext is set to YES
    !----
      allocate (a_x(0),a_y(0),a_yext(0),a_type(0),a_badflag(0))
      Nfinal = 0
      sum_m1 = 0
      sum_m2 = 0
      n12cal = 0
      ntrow3 = 0     
    else
!write (*,*) "npixtot,n12cal,Nbpxfincal,Nadd,Npurge,Noutwin"
!write (*,*)  npixtot,n12cal,Nbpxfincal,Nadd,Npurge,Noutwin
      Nfinal = npixtot - Nbpxfincal - Noutwin - Npurge + Nadd
    endif

    write (errstr,*) Nfinal," final bad pixel sets will be added to BADPIX extension"
    call message(AppMsg,Sparse)
    call message(AppMsg,Verbose,"222   X    Y  TYPE  Y-EXTENT  MODE")
    do n = 1, Nfinal
      write (errstr, "(I5,I5,I5,I8,I8)") a_x(n),a_y(n),a_type(n),a_yext(n),a_badflag(n)            
      call message(AppMsg,Verbose)
    enddo

    if (delete_BADPIX) call deleteBlock(in_events,"BADPIX")

    tab = addTable(in_events,"BADPIX",Nfinal)
    xCol = addColumn(tab,"RAWX",Integer16,"",comment="Raw X coordinate")
    yCol = addColumn(tab,"RAWY",Integer16,"",comment="Raw Y coordinate")
    typeCol = addColumn(tab,"TYPE",Integer16,"",comment="Bad pixel type")
    yextCol = addColumn(tab,"YEXTENT",Integer16,"",comment="Y-direction extent")
    badflagCol = addColumn(tab,"BADFLAG",Integer16,"",comment="Bad pixel flag")
    p_x => int16Data(xCol)
    p_y => int16Data(yCol)
    p_type => int16Data(typeCol)
    p_yext => int16Data(yextCol)    
    p_badflag => int16Data(badflagCol)
    p_x = 0
    p_y = 0
    p_type = 0 
    p_yext = 0
    p_badflag = 0
    p_x = a_x(1:Nfinal)
    p_y = a_y(1:Nfinal)
    p_type = a_type(1:Nfinal) 
    p_yext = a_yext(1:Nfinal)    
    p_badflag = a_badflag(1:Nfinal)
!   call message(AppMsg,Verbose,'pointers are set ...')
    

!------------------------------------------------------------
!	ADD A FEW ATTRIBUTES
!------------------------------------------------------------
    call setAttribute(tab,"TELESCOP",stringAttribute(tab_events,"TELESCOP"), &
                      "Telescope (mission) name")
    call setAttribute(tab,"INSTRUME",stringAttribute(tab_events,"INSTRUME"), &
                      "Instrument name")
    call setAttribute(tab,"CCDID",int8Attribute(tab_events,"CCDID"),"", &
                      "ccd in use")
    call setAttribute(tab,"EXP_ID",int16Attribute(tab_events,"EXP_ID"),"", &
                      "Exposure identifier")
    call setAttribute(tab,"OBS_ID",int16Attribute(tab_events,"OBS_ID"),"", &
                      "Observation identifier")
    if (instrument == "EPN") then 
      call setAttribute(tab,"QUADRANT",int8Attribute(tab_events,"QUADRANT"), &
                      "","quadrant in use")
    endif
    if (mos) then 
      call setAttribute(tab,"CCDNODE",int8Attribute(tab_events,"CCDNODE"), &
                      "","ccd node")
    endif
    call setAttribute(tab,"WINDOWX0",x0,"", &
                      "X coordinate of bottom left corner of window")
    call setAttribute(tab,"WINDOWY0",y0,"", &
                      "Y coordinate of bottom left corner of window")
    call setAttribute(tab,"WINDOWDX",nx,"", &
                      "Size, along x-axis, of window")
    call setAttribute(tab,"WINDOWDY",ny,"", &
                      "Size, along y-axis, of window")

!   call message(AppMsg,Verbose,'few attributes set ...')


!------------------------------------------------------------		
!	ADD HISTORY (can put more information in)
!------------------------------------------------------------		
    history = "bad pixel extension added. Includes: "
    call addHistory(in_events,history)
    if (mode1) then
      write(history,'(a,i4,a)') "Mode1: ",sum_m1," uplinked bad pixel sets (from CCF)"
      call addHistory(in_events,history)
    endif
    if (mode2) then
      write(history,'(a,i4,a)') "Mode2: ",sum_m2," non-uplinked bad pixel sets (from CCF)"
      call addHistory(in_events,history)
    endif
    if (mode3) then
      if (BpxfInCalEver) then 
        write(history,'(a,i4,a,a80)') "Mode3: ",Nfinal-n12cal," badpixfind pixel sets from file ",stringParameter("badpixset")
        call addHistory(in_events,history)
      else 
        write(history,'(a,i4,a,a80)') "Mode3: ",ntrow3," badpixfind pixel sets from file ",stringParameter("badpixset")
        call addHistory(in_events,history)
      endif
    endif
    if (window) then
      if (Noutwin > 0) then
        write(history,'(i4,a)') Noutwin," bad pixel sets out of WINDOW range - removed "
        call addHistory(in_events,history)
      endif
      if (Ncutwin > 0) then
        write(history,'(i4,a)') Ncutwin," bad pixel sets overlapping WINDOW - altered"
        call addHistory(in_events,history)
      endif
    endif

!   call message(AppMsg,Verbose,'now release memory ...')


!------------------------------------------------------------
!	RELEASE OAL
!------------------------------------------------------------
    if (mos .and. oal .and. .NOT.emptyext) then
!     write(errstr,*) 'MOS = ', mos,'  OAL = ', oal,' EMPTY = ',emptyext
!     call message(AppMsg,Verbose)
      call OAL_releaseMemory(OALrawX, OALrawY)
!     call message(AppMsg,Verbose,' x-y done')
      call OAL_releaseMemory(OALType, OALyextent)
!     call message(AppMsg,Verbose,' type ext done')
      call OAL_releaseMemory(OALuplinked)
    endif
!   call message(AppMsg,Verbose,' Done')

!------------------------------------------------------------
!	RELEASE CAL 
!------------------------------------------------------------
    if (.not.emptyext) then
      call CAL_releaseMemory(rawX, rawY)
      call CAL_releaseMemory(Type, yextent)
      call CAL_releaseMemory(uplinked)
    endif
    if (mode3 .and. .not.emptyext) then 
!      call release(col_mode3_rawX)
!      call release(col_mode3_rawY)
!      call release(col_mode3_type)
!      call release(col_mode3_yextent)
    endif
!    call release(xCol)
!    call release(yCol)
!    call release(typeCol)
!    call release(yextCol)
!    call release(badflagCol)
    call release(tab)
    call release(in_events)
!------------------------------------------------------------
!	END PROGRAM
!------------------------------------------------------------
 end subroutine badpix

end module badpix_mod
